﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.CombinedIntegerVectorEncoding;
using HeuristicLab.Encodings.ConditionActionEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Problems.ConditionActionClassification {
  public class ConditionActionClassificationProblemData : ParameterizedNamedItem, IConditionActionProblemData {

    #region default data
    public static string[] defaultVariableNames = new string[] { "a", "b", "c", "d", "e", "f", "g" };
    public static double[,] defaultData = new double[,]{
      {0,0,1,1,0,0,0},
      {0,1,1,1,0,0,0},
      {0,0,1,0,0,0,1},
      {1,0,1,0,1,1,0}
    };
    #endregion

    #region parameter properites
    public IFixedValueParameter<Dataset> DatasetParameter {
      get { return (IFixedValueParameter<Dataset>)Parameters["Dataset"]; }
    }
    public IFixedValueParameter<ReadOnlyCheckedItemList<StringValue>> ConditionVariablesParameter {
      get { return (IFixedValueParameter<ReadOnlyCheckedItemList<StringValue>>)Parameters["ConditionVariables"]; }
    }
    public IFixedValueParameter<ReadOnlyCheckedItemList<StringValue>> ActionVariablesParameter {
      get { return (IFixedValueParameter<ReadOnlyCheckedItemList<StringValue>>)Parameters["ActionVariables"]; }
    }
    public IFixedValueParameter<IntValue> LengthParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters["Length"]; }
    }
    public IFixedValueParameter<IntValue> ActionLengthParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters["ActionLength"]; }
    }
    public IFixedValueParameter<IntMatrix> BoundsParameter {
      get { return (IFixedValueParameter<IntMatrix>)Parameters["Bounds"]; }
    }
    public IFixedValueParameter<IntRange> TrainingPartitionParameter {
      get { return (IFixedValueParameter<IntRange>)Parameters["TrainingPartition"]; }
    }
    public IFixedValueParameter<IntRange> TestPartitionParameter {
      get { return (IFixedValueParameter<IntRange>)Parameters["TestPartition"]; }
    }
    #endregion

    #region properties
    public Dataset Dataset {
      get { return DatasetParameter.Value; }
    }
    public ICheckedItemList<StringValue> ConditionVariables {
      get { return ConditionVariablesParameter.Value; }
    }
    public ICheckedItemList<StringValue> ActionVariables {
      get { return ActionVariablesParameter.Value; }
    }
    public IEnumerable<string> AllowedConditionVariables {
      get { return ConditionVariables.CheckedItems.Select(x => x.Value.Value); }
    }
    public IEnumerable<string> AllowedActionVariables {
      get { return ActionVariables.CheckedItems.Select(x => x.Value.Value); }
    }
    public IntRange TrainingPartition {
      get { return TrainingPartitionParameter.Value; }
    }
    public IntRange TestPartition {
      get { return TestPartitionParameter.Value; }
    }
    public IEnumerable<int> TrainingIndices {
      get {
        return Enumerable.Range(TrainingPartition.Start, Math.Max(0, TrainingPartition.End - TrainingPartition.Start))
                         .Where(IsTrainingSample);
      }
    }
    public IEnumerable<int> TestIndices {
      get {
        return Enumerable.Range(TestPartition.Start, Math.Max(0, TestPartition.End - TestPartition.Start))
           .Where(IsTestSample);
      }
    }
    public bool IsTrainingSample(int index) {
      return index >= 0 && index < Dataset.Rows &&
        TrainingPartition.Start <= index && index < TrainingPartition.End &&
        (index < TestPartition.Start || TestPartition.End <= index);
    }
    public bool IsTestSample(int index) {
      return index >= 0 && index < Dataset.Rows &&
             TestPartition.Start <= index && index < TestPartition.End;
    }
    #endregion

    private IDictionary<int, IClassifier> fetchClassifiersCache = new Dictionary<int, IClassifier>();


    [StorableConstructor]
    protected ConditionActionClassificationProblemData(bool deserializing) : base(deserializing) { }
    protected ConditionActionClassificationProblemData(ConditionActionClassificationProblemData original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ConditionActionClassificationProblemData(this, cloner);
    }

    public ConditionActionClassificationProblemData(Dataset dataset, IEnumerable<string> allowedConditionVariables, IEnumerable<string> allowedActionVariables) {
      if (dataset == null) throw new ArgumentNullException("The dataset must not be null.");
      if (allowedActionVariables == null) throw new ArgumentNullException("The allowedActionVariables must not be null.");
      if (allowedConditionVariables == null) throw new ArgumentNullException("The allowedActionVariables must not be null.");

      if (allowedActionVariables.Except(dataset.DoubleVariables).Any())
        throw new ArgumentException("All allowed action variables must be present in the dataset and of type double.");
      if (allowedConditionVariables.Except(dataset.DoubleVariables).Any())
        throw new ArgumentException("All allowed condition variables must be present in the dataset and of type double.");

      var actionVariables = new CheckedItemList<StringValue>(dataset.DoubleVariables.Select(x => new StringValue(x)));
      var conditionVariables = new CheckedItemList<StringValue>(actionVariables);
      foreach (StringValue x in actionVariables) {
        actionVariables.SetItemCheckedState(x, allowedActionVariables.Contains(x.Value));
        conditionVariables.SetItemCheckedState(x, allowedConditionVariables.Contains(x.Value));
      }

      int trainingPartitionStart = 0;
      int trainingPartitionEnd = dataset.Rows / 2;
      int testPartitionStart = dataset.Rows / 2;
      int testPartitionEnd = dataset.Rows;

      Parameters.Add(new FixedValueParameter<Dataset>("Dataset", "", dataset));
      Parameters.Add(new FixedValueParameter<ReadOnlyCheckedItemList<StringValue>>("ActionVariables", "", actionVariables.AsReadOnly()));
      Parameters.Add(new FixedValueParameter<ReadOnlyCheckedItemList<StringValue>>("ConditionVariables", "", conditionVariables.AsReadOnly()));
      Parameters.Add(new FixedValueParameter<IntValue>("Length", "", new IntValue(allowedConditionVariables.Count() + allowedActionVariables.Count())));
      Parameters.Add(new FixedValueParameter<IntValue>("ActionLength", "", new IntValue(allowedActionVariables.Count())));
      Parameters.Add(new FixedValueParameter<IntMatrix>("Bounds", "", GetBoundsMatrix(dataset, allowedConditionVariables, allowedActionVariables)));
      Parameters.Add(new FixedValueParameter<IntRange>("TrainingPartition", "", new IntRange(trainingPartitionStart, trainingPartitionEnd)));
      Parameters.Add(new FixedValueParameter<IntRange>("TestPartition", "", new IntRange(testPartitionStart, testPartitionEnd)));

      ((ValueParameter<Dataset>)DatasetParameter).ReactOnValueToStringChangedAndValueItemImageChanged = false;
    }

    private IntMatrix GetBoundsMatrix(Dataset dataset, IEnumerable<string> conditionVariables, IEnumerable<string> actionVariables) {
      IntMatrix bounds = new IntMatrix(conditionVariables.Count() + actionVariables.Count(), 2);
      int index = 0;
      foreach (var variable in conditionVariables) {
        var values = dataset.GetDoubleValues(variable);
        bounds[index, 0] = (int)values.Min();
        bounds[index, 1] = (int)values.Max() + 2;
        index++;
      }
      foreach (var variable in actionVariables) {
        var values = dataset.GetDoubleValues(variable);
        bounds[index, 0] = (int)values.Min();
        bounds[index, 1] = (int)values.Max() + 1;
        index++;
      }
      return bounds;
    }

    public event EventHandler Changed;
    protected virtual void OnChanged() {
      var listeners = Changed;
      if (listeners != null) listeners(this, EventArgs.Empty);
    }

    public IntValue Length {
      get { return LengthParameter.Value; }
    }

    public IntValue ActionLength {
      get { return ActionLengthParameter.Value; }
    }

    public IntMatrix Bounds {
      get { return BoundsParameter.Value; }
    }

    public IEnumerable<IClassifier> FetchClassifier(IEnumerable<int> rows) {
      foreach (var row in rows) {
        yield return FetchClassifier(row);
      }
    }

    public IClassifier FetchClassifier(int rowNumber) {
      if (!fetchClassifiersCache.ContainsKey(rowNumber)) {
        int[] elements = new int[Length.Value];
        var variableNamesList = Dataset.VariableNames.ToList();
        int elementIndex = 0;
        for (int i = 0; i < variableNamesList.Count; i++) {
          if (AllowedConditionVariables.Contains(variableNamesList[i])) {
            elements[elementIndex] = int.Parse(Dataset.GetValue(rowNumber, i));
            elementIndex++;
          }
        }
        for (int i = 0; i < variableNamesList.Count; i++) {
          if (AllowedActionVariables.Contains(variableNamesList[i])) {
            elements[elementIndex] = int.Parse(Dataset.GetValue(rowNumber, i));
            elementIndex++;
          }
        }
        if (elementIndex != Length.Value) {
          throw new ArgumentException("Length of classifier is not equal to the number of allowed condition + action variables.");
        }
        fetchClassifiersCache.Add(rowNumber, new CombinedIntegerVector(elements, ActionLengthParameter.Value.Value, BoundsParameter.Value));
      }
      return fetchClassifiersCache[rowNumber];
    }
  }
}
